//////////////////////////////////////////////////////////////////////////////// 
/// Plainamp, Open source Winamp core
/// 
/// Copyright  2005  Sebastian Pipping <webmaster@hartwork.org>
/// 
/// -->  http://www.hartwork.org
/// 
/// This source code is released under the GNU General Public License (GPL).
/// See GPL.txt for details. Any non-GPL usage is strictly forbidden.
////////////////////////////////////////////////////////////////////////////////


#include "OutputPlugin.h"
#include "Main.h"
#include "Unicode.h"
#include "Console.h"



vector <OutputPlugin *> output_plugins; // extern
OutputPlugin ** active_output_plugins = NULL; // extern
int active_output_count = 0; // extern



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
OutputPlugin::OutputPlugin( TCHAR * szDllpath, bool bKeepLoaded ) : Plugin( szDllpath )
{
	bActive      = false;
	iArrayIndex  = -1;
	plugin       = NULL;
	
	if( !Load() ) 
	{
		return;
	}

////////////////////////////////////////////////////////////////////////////////
	// TODO
	if( !_tcscmp( GetFilename(), TEXT( "out_ds.dll" ) ) )
	{
		Start();
	}
	else
	{
		if( !bKeepLoaded )
		{
			Unload();
		}
	}
////////////////////////////////////////////////////////////////////////////////
	
	
	output_plugins.push_back( this );
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::Load()
{
	if( IsLoaded() ) return true;

	// (1) Load DLL
	hDLL = LoadLibrary( GetFullpath() );
	if( !hDLL ) return false;

	// (2) Find export
	WINAMP_OUTPUT_GETTER winampGetOutModule =
		( WINAMP_OUTPUT_GETTER )GetProcAddress(	hDLL, "winampGetOutModule" );
	if( winampGetOutModule == NULL )
	{
		FreeLibrary( hDLL );
		hDLL = NULL;
		return false;
	}

	// (3) Get module
	plugin = winampGetOutModule();
	if( !plugin )
	{
		FreeLibrary( hDLL );
		hDLL = NULL;
		return false;
	}

	// (4) Process module
	plugin->hDllInstance  = hDLL;
	plugin->hMainWindow   = WindowMain;

	if( !szName )
	{
		// Note:  The prefix is not removed to hide their
		//        origin at Nullsoft! It just reads easier.
		if( !strnicmp( plugin->description, "nullsoft ", 9 ) )
		{
			plugin->description += 9;
		}
		iNameLen = ( int )strlen( plugin->description );
		szName = new TCHAR[ iNameLen + 1 ];
		ToTchar( szName, plugin->description, iNameLen );
		szName[ iNameLen ] = TEXT( '\0' );
	}


	TCHAR szBuffer[ 5000 ];
	_stprintf( szBuffer, TEXT( "Loading <%s>, %s" ), GetFilename(), szName );
	Console::Append( szBuffer );
	Console::Append( TEXT( " " ) );
	
	if( plugin->Init )
	{
		plugin->Init();
	}

	return true;
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::Unload()
{
	if( !IsLoaded() ) return true;
	if( bActive ) return false;

	TCHAR szBuffer[ 5000 ];
	_stprintf( szBuffer, TEXT( "Unloading <%s>" ), GetFilename() );
	Console::Append( szBuffer );
	Console::Append( TEXT( " " ) );

	// Quit
	if( plugin )
	{
		// if( plugin->Close ) plugin->Close();
		if( plugin->Quit ) plugin->Quit();
		plugin = NULL;
	}

	FreeLibrary( hDLL );
	hDLL = NULL;

	return true;
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::About( HWND hParent )
{
	if( !IsLoaded() ) return false;
	if( !plugin ) return false;
	if( !plugin->About ) return false;
	
	plugin->About( hParent );
	
	return true;
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::Config( HWND hParent )
{
	if( !IsLoaded() ) return false;
	if( !plugin ) return false;
	if( !plugin->Config ) return false;
	
	plugin->Config( hParent );
	
	return true;
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::Start()
{
	if( !IsLoaded() ) return false;
	if( bActive ) return true;

	if( active_output_count )
	{
		active_output_plugins = ( OutputPlugin ** )realloc( active_output_plugins, ( active_output_count + 1 ) * sizeof( OutputPlugin * ) );
		active_output_plugins[ active_output_count ] = this;
		iArrayIndex = active_output_count;
		active_output_count++;
	}
	else
	{
		active_output_plugins = ( OutputPlugin ** )malloc( sizeof( OutputPlugin * ) );
		active_output_plugins[ 0 ] = this;
		iArrayIndex = 0;
		active_output_count = 1;
	}
	
	TCHAR szBuffer[ 5000 ];
	_stprintf( szBuffer, TEXT( "Output plugin <%s> activated" ), GetFilename() );
	Console::Append( szBuffer );
	Console::Append( TEXT( " " ) );
	
	bActive = true;

	return true;
}



////////////////////////////////////////////////////////////////////////////////
///
////////////////////////////////////////////////////////////////////////////////
bool OutputPlugin::Stop()
{
	if( !IsLoaded() ) return true;
	if( !bActive ) return true;

	const int iMaxIndex = active_output_count - 1;
	if( iArrayIndex < iMaxIndex )
	{
		active_output_plugins[ iArrayIndex ] = active_output_plugins[ iMaxIndex ];
		active_output_plugins[ iArrayIndex ]->iArrayIndex = iArrayIndex;
	}
	iArrayIndex = -1;
	active_output_count--;
	
	// TODO Flush?
	// TODO Close?
	
	TCHAR szBuffer[ 5000 ];
	_stprintf( szBuffer, TEXT( "Output plugin <%s> deactivated" ), GetFilename() );
	Console::Append( szBuffer );
	Console::Append( TEXT( " " ) );
	
	bActive = false;

	return true;
}
